@@ -16,6 +16,7 @@
 """
 Key manager implementation for Barbican
 """
+import calendar
 import time
 
 from cryptography.hazmat import backends
@@ -40,8 +41,10 @@ from castellan.openstack.common import _i18n as u
 
 from barbicanclient import client as barbican_client
 from barbicanclient import exceptions as barbican_exceptions
+from oslo_utils import timeutils
 from six.moves import urllib
 
+
 barbican_opts = [
     cfg.StrOpt('barbican_endpoint',
                help='Use this endpoint to connect to Barbican, for example: '
@@ -90,8 +93,8 @@ class BarbicanKeyManager(key_manager.KeyManager):
         :param context: the user context for authentication
         :return: a Barbican Client object
         :raises Forbidden: if the context is None
-        :raises KeyManagerError: if context is missing tenant or
-                                 tenant is None
+        :raises KeyManagerError: if context is missing tenant or tenant is
+                                 None or error occurs while creating client
         """
 
         # Confirm context is provided, if not raise forbidden
@@ -100,17 +103,10 @@ class BarbicanKeyManager(key_manager.KeyManager):
             LOG.error(msg)
             raise exception.Forbidden(msg)
 
-        if not hasattr(context, 'tenant') or context.tenant is None:
-            msg = u._("Unable to create Barbican Client without tenant "
-                      "attribute in context object.")
-            LOG.error(msg)
-            raise exception.KeyManagerError(msg)
-
         if self._barbican_client and self._current_context == context:
             return self._barbican_client
 
         try:
-            self._current_context = context
             auth = self._get_keystone_auth(context)
             sess = session.Session(auth=auth)
 
@@ -118,10 +114,11 @@ class BarbicanKeyManager(key_manager.KeyManager):
             self._barbican_client = barbican_client.Client(
                 session=sess,
                 endpoint=self._barbican_endpoint)
+            self._current_context = context
 
         except Exception as e:
-            with excutils.save_and_reraise_exception():
-                LOG.error(u._LE("Error creating Barbican client: %s"), e)
+            LOG.error(u._LE("Error creating Barbican client: %s"), e)
+            raise exception.KeyManagerError(reason=e)
 
         self._base_url = self._create_base_url(auth,
                                                sess,
@@ -130,14 +127,48 @@ class BarbicanKeyManager(key_manager.KeyManager):
         return self._barbican_client
 
     def _get_keystone_auth(self, context):
-        # TODO(kfarr): support keystone v2
-        auth = identity.v3.Token(
-            auth_url=self.conf.barbican.auth_endpoint,
-            token=context.auth_token,
-            project_id=context.tenant,
-            domain_id=context.user_domain,
-            project_domain_id=context.project_domain)
-        return auth
+        auth_url = self.conf.barbican.auth_endpoint
+
+        if context.__class__.__name__ is 'KeystonePassword':
+            return identity.v3.Password(
+                auth_url=auth_url,
+                username=context.username,
+                password=context.password,
+                user_id=context.user_id,
+                user_domain_id=context.user_domain_id,
+                user_domain_name=context.user_domain_name,
+                trust_id=context.trust_id,
+                domain_id=context.domain_id,
+                domain_name=context.domain_name,
+                project_id=context.project_id,
+                project_name=context.project_name,
+                project_domain_id=context.project_domain_id,
+                project_domain_name=context.project_domain_name,
+                reauthenticate=context.reauthenticate)
+        elif context.__class__.__name__ is 'KeystoneToken':
+            return identity.v3.Token(
+                auth_url=auth_url,
+                token=context.token,
+                trust_id=context.trust_id,
+                domain_id=context.domain_id,
+                domain_name=context.domain_name,
+                project_id=context.project_id,
+                project_name=context.project_name,
+                project_domain_id=context.project_domain_id,
+                project_domain_name=context.project_domain_name,
+                reauthenticate=context.reauthenticate)
+        # this will be kept for oslo.context compatibility until
+        # projects begin to use utils.credential_factory
+        elif context.__class__.__name__ is 'RequestContext':
+            return identity.v3.Token(
+                auth_url=auth_url,
+                token=context.auth_token,
+                project_id=context.tenant)
+        else:
+            msg = "context must be of type KeystonePassword, KeystoneToken, "
+            "or RequestContext."
+            LOG.error(msg)
+            raise exception.Forbidden(reason=msg)
 
     def _get_barbican_endpoint(self, auth, sess):
         if self.conf.barbican.barbican_endpoint:
@@ -158,7 +189,7 @@ class BarbicanKeyManager(key_manager.KeyManager):
                 msg = u._LE(
                     "Could not find discovery information for %s") % endpoint
                 LOG.error(msg)
-                raise exception.KeyManagerError(msg)
+                raise exception.KeyManagerError(reason=msg)
             latest_version = raw_data[-1]
             api_version = latest_version.get('id')
 
@@ -166,23 +197,24 @@ class BarbicanKeyManager(key_manager.KeyManager):
             endpoint, api_version)
         return base_url
 
-    def create_key(self, context, algorithm, length, expiration=None):
+    def create_key(self, context, algorithm, length,
+                   expiration=None, name=None):
         """Creates a symmetric key.
 
         :param context: contains information of the user and the environment
                         for the request (castellan/context.py)
         :param algorithm: the algorithm associated with the secret
         :param length: the bit length of the secret
+        :param name: the name of the key
         :param expiration: the date the key will expire
         :return: the UUID of the new key
-        :raises HTTPAuthError: if key creation fails with 401
-        :raises HTTPClientError: if key creation failes with 4xx
-        :raises HTTPServerError: if key creation fails with 5xx
+        :raises KeyManagerError: if key creation fails
         """
         barbican_client = self._get_barbican_client(context)
 
         try:
             key_order = barbican_client.orders.create_key(
+                name=name,
                 algorithm=algorithm,
                 bit_length=length,
                 expiration=expiration)
@@ -192,22 +224,22 @@ class BarbicanKeyManager(key_manager.KeyManager):
         except (barbican_exceptions.HTTPAuthError,
                 barbican_exceptions.HTTPClientError,
                 barbican_exceptions.HTTPServerError) as e:
-            with excutils.save_and_reraise_exception():
-                LOG.error(u._LE("Error creating key: %s"), e)
+            LOG.error(u._LE("Error creating key: %s"), e)
+            raise exception.KeyManagerError(reason=e)
 
-    def create_key_pair(self, context, algorithm, length, expiration=None):
+    def create_key_pair(self, context, algorithm, length,
+                        expiration=None, name=None):
         """Creates an asymmetric key pair.
 
         :param context: contains information of the user and the environment
                         for the request (castellan/context.py)
         :param algorithm: the algorithm associated with the secret
         :param length: the bit length of the secret
+        :param name: the name of the key
         :param expiration: the date the key will expire
         :return: the UUIDs of the new key, in the order (private, public)
         :raises NotImplementedError: until implemented
-        :raises HTTPAuthError: if key creation fails with 401
-        :raises HTTPClientError: if key creation failes with 4xx
-        :raises HTTPServerError: if key creation fails with 5xx
+        :raises KeyManagerError: if key pair creation fails
         """
         barbican_client = self._get_barbican_client(context)
 
@@ -215,6 +247,7 @@ class BarbicanKeyManager(key_manager.KeyManager):
             key_pair_order = barbican_client.orders.create_asymmetric(
                 algorithm=algorithm,
                 bit_length=length,
+                name=name,
                 expiration=expiration)
 
             order_ref = key_pair_order.submit()
@@ -229,11 +262,13 @@ class BarbicanKeyManager(key_manager.KeyManager):
         except (barbican_exceptions.HTTPAuthError,
                 barbican_exceptions.HTTPClientError,
                 barbican_exceptions.HTTPServerError) as e:
-            with excutils.save_and_reraise_exception():
-                LOG.error(u._LE("Error creating key pair: %s"), e)
+            LOG.error(u._LE("Error creating key pair: %s"), e)
+            raise exception.KeyManagerError(reason=e)
 
     def _get_barbican_object(self, barbican_client, managed_object):
         """Converts the Castellan managed_object to a Barbican secret."""
+        name = getattr(managed_object, 'name', None)
+
         try:
             algorithm = managed_object.algorithm
             bit_length = managed_object.bit_length
@@ -248,6 +283,7 @@ class BarbicanKeyManager(key_manager.KeyManager):
         secret = barbican_client.secrets.create(payload=payload,
                                                 algorithm=algorithm,
                                                 bit_length=bit_length,
+                                                name=name,
                                                 secret_type=secret_type)
         return secret
 
@@ -287,14 +323,12 @@ class BarbicanKeyManager(key_manager.KeyManager):
 
         :param context: contains information of the user and the environment
             for the request (castellan/context.py)
-        :param managed_object: the unencrypted secret data. Known as "payload"
-            to the barbicanclient api
+        :param managed_object: a secret object with unencrypted payload.
+            Known as "secret" to the barbicanclient api
         :param expiration: the expiration time of the secret in ISO 8601
             format
         :returns: the UUID of the stored object
-        :raises HTTPAuthError: if object creation fails with 401
-        :raises HTTPClientError: if object creation failes with 4xx
-        :raises HTTPServerError: if object creation fails with 5xx
+        :raises KeyManagerError: if object store fails
         """
         barbican_client = self._get_barbican_client(context)
 
@@ -307,22 +341,22 @@ class BarbicanKeyManager(key_manager.KeyManager):
         except (barbican_exceptions.HTTPAuthError,
                 barbican_exceptions.HTTPClientError,
                 barbican_exceptions.HTTPServerError) as e:
-            with excutils.save_and_reraise_exception():
-                LOG.error(u._LE("Error storing object: %s"), e)
+            LOG.error(u._LE("Error storing object: %s"), e)
+            raise exception.KeyManagerError(reason=e)
 
-    def _create_secret_ref(self, key_id):
+    def _create_secret_ref(self, object_id):
         """Creates the URL required for accessing a secret.
 
-        :param key_id: the UUID of the key to copy
+        :param object_id: the UUID of the key to copy
         :return: the URL of the requested secret
         """
-        if not key_id:
+        if not object_id:
             msg = "Key ID is None"
-            raise exception.KeyManagerError(msg)
+            raise exception.KeyManagerError(reason=msg)
         base_url = self._base_url
         if base_url[-1] != '/':
             base_url += '/'
-        return urllib.parse.urljoin(base_url, "secrets/" + key_id)
+        return urllib.parse.urljoin(base_url, "secrets/" + object_id)
 
     def _get_active_order(self, barbican_client, order_ref):
         """Returns the order when it is active.
@@ -330,17 +364,26 @@ class BarbicanKeyManager(key_manager.KeyManager):
         Barbican key creation is done asynchronously, so this loop continues
         checking until the order is active or a timeout occurs.
         """
-        active = u'ACTIVE'
+        active_status = u'ACTIVE'
+        error_status = u'ERROR'
         number_of_retries = self.conf.barbican.number_of_retries
         retry_delay = self.conf.barbican.retry_delay
         order = barbican_client.orders.get(order_ref)
         time.sleep(.25)
         for n in range(number_of_retries):
-            if order.status != active:
+            if order.status == error_status:
+                kwargs = {"status": error_status,
+                          "code": order.error_status_code,
+                          "reason": order.error_reason}
+                msg = u._LE("Order is in %(status)s status - status code: "
+                            "%(code)s, status reason: %(reason)s") % kwargs
+                LOG.error(msg)
+                raise exception.KeyManagerError(reason=msg)
+            if order.status != active_status:
                 kwargs = {'attempt': n,
                           'total': number_of_retries,
                           'status': order.status,
-                          'active': active,
+                          'active': active_status,
                           'delay': retry_delay}
                 msg = u._LI("Retry attempt #%(attempt)i out of %(total)i: "
                             "Order status is '%(status)s'. Waiting for "
@@ -352,11 +395,11 @@ class BarbicanKeyManager(key_manager.KeyManager):
             else:
                 return order
         msg = u._LE("Exceeded retries: Failed to find '%(active)s' status "
-                    "within %(num_retries)i retries") % {'active': active,
-                                                         'num_retries':
-                                                         number_of_retries}
+                    "within %(num_retries)i retries") % {
+            'active': active_status,
+            'num_retries': number_of_retries}
         LOG.error(msg)
-        raise exception.KeyManagerError(msg)
+        raise exception.KeyManagerError(reason=msg)
 
     def _retrieve_secret_uuid(self, secret_ref):
         """Retrieves the UUID of the secret from the secret_ref.
@@ -418,19 +461,29 @@ class BarbicanKeyManager(key_manager.KeyManager):
 
         secret_data = self._get_secret_data(secret)
 
+        # convert created ISO8601 in Barbican to POSIX
+        if secret.created:
+            time_stamp = timeutils.parse_isotime(
+                str(secret.created)).timetuple()
+            created = calendar.timegm(time_stamp)
+
         if issubclass(secret_type, key_base_class.Key):
             return secret_type(secret.algorithm,
                                secret.bit_length,
-                               secret_data)
+                               secret_data,
+                               secret.name,
+                               created)
         else:
-            return secret_type(secret_data)
+            return secret_type(secret_data,
+                               secret.name,
+                               created)
 
-    def _get_secret(self, context, key_id):
+    def _get_secret(self, context, object_id):
         """Returns the metadata of the secret.
 
         :param context: contains information of the user and the environment
                         for the request (castellan/context.py)
-        :param key_id: UUID of the secret
+        :param object_id: UUID of the secret
         :return: the secret's metadata
         :raises HTTPAuthError: if object retrieval fails with 401
         :raises HTTPClientError: if object retrieval fails with 4xx
@@ -440,7 +493,7 @@ class BarbicanKeyManager(key_manager.KeyManager):
         barbican_client = self._get_barbican_client(context)
 
         try:
-            secret_ref = self._create_secret_ref(key_id)
+            secret_ref = self._create_secret_ref(object_id)
             return barbican_client.secrets.get(secret_ref)
         except (barbican_exceptions.HTTPAuthError,
                 barbican_exceptions.HTTPClientError,
@@ -448,6 +501,13 @@ class BarbicanKeyManager(key_manager.KeyManager):
             with excutils.save_and_reraise_exception():
                 LOG.error(u._LE("Error getting secret metadata: %s"), e)
 
+    def _is_secret_not_found_error(self, error):
+        if (isinstance(error, barbican_exceptions.HTTPClientError) and
+                error.status_code == 404):
+            return True
+        else:
+            return False
+
     def get(self, context, managed_object_id):
         """Retrieves the specified managed object.
 
@@ -457,9 +517,8 @@ class BarbicanKeyManager(key_manager.KeyManager):
                         for the request (castellan/context.py)
         :param managed_object_id: the UUID of the object to retrieve
         :return: SymmetricKey representation of the key
-        :raises HTTPAuthError: if object retrieval fails with 401
-        :raises HTTPClientError: if object retrieval fails with 4xx
-        :raises HTTPServerError: if object retrieval fails with 5xx
+        :raises KeyManagerError: if object retrieval fails
+        :raises ManagedObjectNotFoundError: if object not found
         """
         try:
             secret = self._get_secret(context, managed_object_id)
@@ -467,8 +526,12 @@ class BarbicanKeyManager(key_manager.KeyManager):
         except (barbican_exceptions.HTTPAuthError,
                 barbican_exceptions.HTTPClientError,
                 barbican_exceptions.HTTPServerError) as e:
-            with excutils.save_and_reraise_exception():
-                LOG.error(u._LE("Error getting object: %s"), e)
+            LOG.error(u._LE("Error retrieving object: %s"), e)
+            if self._is_secret_not_found_error(e):
+                raise exception.ManagedObjectNotFoundError(
+                    uuid=managed_object_id)
+            else:
+                raise exception.KeyManagerError(reason=e)
 
     def delete(self, context, managed_object_id):
         """Deletes the specified managed object.
@@ -476,9 +539,8 @@ class BarbicanKeyManager(key_manager.KeyManager):
         :param context: contains information of the user and the environment
                      for the request (castellan/context.py)
         :param managed_object_id: the UUID of the object to delete
-        :raises HTTPAuthError: if key deletion fails with 401
-        :raises HTTPClientError: if key deletion fails with 4xx
-        :raises HTTPServerError: if key deletion fails with 5xx
+        :raises KeyManagerError: if key deletion fails
+        :raises ManagedObjectNotFoundError: if the object could not be found
         """
         barbican_client = self._get_barbican_client(context)
 
@@ -488,5 +550,9 @@ class BarbicanKeyManager(key_manager.KeyManager):
         except (barbican_exceptions.HTTPAuthError,
                 barbican_exceptions.HTTPClientError,
                 barbican_exceptions.HTTPServerError) as e:
-            with excutils.save_and_reraise_exception():
-                LOG.error(u._LE("Error deleting object: %s"), e)
+            LOG.error(u._LE("Error deleting object: %s"), e)
+            if self._is_secret_not_found_error(e):
+                raise exception.ManagedObjectNotFoundError(
+                    uuid=managed_object_id)
+            else:
+                raise exception.KeyManagerError(reason=e)
